﻿using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Runtime.InteropServices;
using MAPI;
using Application = Outlook.Application;
using Exception = System.Exception;

namespace OutlookSample
{
  public class MailAddressMapper
  {
    #region Const
    private const int PR_MAIL = 0x39FE001E; //972947486;
    #endregion

    #region Declarations
    private volatile Dictionary<int, string> Collection = new Dictionary<int, string>();
    private Session MapiSession;
    private static MailAddressMapper Current = new MailAddressMapper();
    #endregion

    #region Constructors
    /// <summary>
    /// 
    /// </summary>
    private MailAddressMapper() { }
    #endregion

    #region Methods
    /// <summary>
    /// 
    /// </summary>
    /// <param name="mailAddress"></param>
    /// <param name="outlookApp"></param>
    /// <returns></returns>
    public static string Find(string mailAddress, Application outlookApp)
    {
      try
      {
        if (string.IsNullOrEmpty(mailAddress))
          return null;

        var hash = mailAddress.GetHashCode();

        if (Current.Collection.Keys.Contains(hash))
        {
          mailAddress = Current.Collection[hash];
        }
        else
        {
          // get mapi mail address
          if (mailAddress.StartsWith("/O=")
            || mailAddress.StartsWith("/o=")
            || mailAddress.StartsWith("/0="))
          {
            mailAddress = Current.GetSmtpAddress(mailAddress, outlookApp);
          }

          Current.Collection.Add(hash, mailAddress);
        }
      }
      catch (Exception) { }

      return mailAddress;
    }
    /// <summary>
    /// 
    /// </summary>
    private void InitMapiSession()
    {
      try
      {
        this.MapiSession = new MAPI.Session();
        this.MapiSession.Logon("", "", false, false, null, Missing.Value, Missing.Value);
      }
      catch (Exception) { }
    }
    /// <summary>
    /// 
    /// </summary>
    public static void Dispose()
    {
      try
      {
        if (Current.MapiSession != null)
          Current.MapiSession.Logoff();

        SafeRelease(Current.MapiSession);
      }
      catch (Exception) { }
      finally
      {
        Current.Collection.Clear();
        Current.MapiSession = null;
      }
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="mailAddress"></param>
    /// <param name="outlookApp"></param>
    /// <returns></returns>
    private string GetSmtpAddress(string mailAddress, Application outlookApp)
    {
      Outlook.Recipient recipient = null;
      Outlook.AddressEntry addressEntry = null;
      MAPI.AddressEntry mapiAddressEntry = null;

      try
      {
        if (this.MapiSession == null)
          this.InitMapiSession();

        recipient = outlookApp.Session.CreateRecipient(mailAddress);

        if (recipient != null)
        {
          addressEntry = recipient.AddressEntry;

          if (addressEntry != null)
          {
            mapiAddressEntry = MapiSession.GetAddressEntry(addressEntry.ID) as MAPI.AddressEntry;
            if (mapiAddressEntry != null)
            {
              var newMailAddress = RecipientToSMTPAddress(mapiAddressEntry);
              if (!string.IsNullOrEmpty(newMailAddress))
                mailAddress = newMailAddress;
            }
          }
        }
      }
      catch (Exception) { }
      finally
      {
        SafeRelease(recipient);
        SafeRelease(addressEntry);
        SafeRelease(mapiAddressEntry);
      }

      return mailAddress;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="addressEntry"></param>
    /// <returns></returns>
    private string RecipientToSMTPAddress(MAPI.AddressEntry addressEntry)
    {
      string mailAddress = null;

      MAPI.Fields fields = null;
      MAPI.Field field = null;

      try
      {
        fields = addressEntry.Fields as MAPI.Fields;
        field = fields.get_Item(PR_MAIL, System.Reflection.Missing.Value) as MAPI.Field;
        if (field != null && field.Value != null)
          mailAddress = field.Value as string;
      }
      catch (Exception) { }
      finally
      {
        SafeRelease(fields);
        SafeRelease(field);
      }

      return mailAddress;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="obj"></param>
    private static void SafeRelease(object obj)
    {
      if (obj == null || !Marshal.IsComObject(obj))
        return;

      Marshal.ReleaseComObject(obj);
    }
    #endregion
  }
}

