﻿using System;
using System.Reflection;
using System.Runtime.InteropServices;
using AddinExpress.Outlook;
using Outlook;
using Exception = System.Exception;
using System.Globalization;
using System.Threading.Tasks;

namespace OutlookSample
{
  public class OutlookClient
  {
    #region Enums
    /// <summary>
    /// 
    /// </summary>
    private enum OlBodyFormat
    {
      OlFormatUnspecified = 0,
      OlFormatPlain = 1,
      OlFormatHtml = 2,
      OlFormatRichText = 3,
    }
    #endregion

    #region Const
    /// <summary>
    /// 
    /// </summary>
    public static CultureInfo CiUs = CultureInfo.CreateSpecificCulture("en-us");
    #endregion

    #region Declarations
    private Folders _Root;
    #endregion

    #region Properties
    /// <summary>
    /// 
    /// </summary>
    public Folders MailFolders
    {
      get
      {
        if (_Root == null)
        {
          _Root = CreateFolders(null, null);
        }

        return _Root;
      }
    }
    #endregion

    #region Methods
    /// <summary>
    /// 
    /// </summary>
    /// <param name="type"></param>
    /// <returns></returns>
    public Folder CreateFolderByType(OlDefaultFolders type)
    {
      Folder folder = null;
      Application outlookApp = null;
      MAPIFolder mapiFolder = null;

      try
      {
        try
        {
          outlookApp = Marshal.GetActiveObject("Outlook.Application") as Application;
        }
        catch (Exception)
        {
          outlookApp = new Application();
        }

        var secMan = BeginSecurity(outlookApp); // first error, set breakpoint here!
        mapiFolder = outlookApp.Session.GetDefaultFolder(type);

        if (mapiFolder != null)
          folder = CreateFolderByID(mapiFolder.EntryID, mapiFolder.StoreID, false);

        EndSecurity(secMan, outlookApp);
      }
      catch (Exception ex) { throw ex; }
      finally
      {
        SafeRelease(mapiFolder);
        SafeRelease(outlookApp);
      }

      return folder;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="entryID"></param>
    /// <param name="storeID"></param>
    /// <returns></returns>
    public Mails CreateMailsFromFolder(string entryID, string storeID)
    {
      return CreateMailsFromFolder(entryID, storeID, null);
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="folderID"></param>
    /// <param name="storeID"></param>
    /// <returns></returns>
    public Folders CreateFolders(string folderID, string storeID)
    {
      Application outlookApp = null;
      MAPIFolder mapiRootFolder = null;
      var rootFolders = new Folders(this);

      try
      {
        try
        {
          outlookApp = Marshal.GetActiveObject("Outlook.Application") as Application;
        }
        catch (Exception)
        {
          outlookApp = new Application();
        }

        var secMan = BeginSecurity(outlookApp);

        // clear
        if (!string.IsNullOrEmpty(folderID) && !string.IsNullOrEmpty(storeID))
        {
          if (outlookApp != null)
            mapiRootFolder = outlookApp.Session.GetFolderFromID(folderID, storeID);

          if (mapiRootFolder != null)
          {
            Parallel.For(1, mapiRootFolder.Folders.Count + 1, i =>
            {
              var f = mapiRootFolder.Folders.Item(i);
              var folder = new Folder(this)
              {
                DisplayName = f.Name,
                Name = f.Name,
                ID = string.Format("{0}*{1}", f.EntryID, f.StoreID),
              };

              rootFolders.Add(folder);
            });
          }
        }
        else if (outlookApp != null)
        {
          Parallel.For(1, outlookApp.Session.Folders.Count + 1, i =>
          {
            var f = outlookApp.Session.Folders.Item(i);
            var folder = new Folder(this)
            {
              DisplayName = f.Name,
              Name = f.Name,
              ID = string.Format("{0}*{1}", f.EntryID, f.StoreID),
            };
            rootFolders.Add(folder);
          });
        }

        EndSecurity(secMan, outlookApp);
      }
      catch (Exception) { }
      finally
      {
        SafeRelease(mapiRootFolder);
        SafeRelease(outlookApp);
      }

      return rootFolders;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="entryID"></param>
    /// <param name="storeID"></param>
    /// <param name="filter"></param>
    /// <returns></returns>
    public Mails CreateMailsFromFolder(string entryID, string storeID, string filter)
    {
      var mails = new Mails(this, entryID, storeID, filter);

      Application outlookApp = null;
      MAPIFolder mapiRootFolder = null;
      Items items = null;

      try
      {
        try
        {
          outlookApp = Marshal.GetActiveObject("Outlook.Application") as Application;
        }
        catch (Exception)
        {
          outlookApp = new Application();
        }

        var secMan = BeginSecurity(outlookApp);

        if (!string.IsNullOrEmpty(entryID) && !string.IsNullOrEmpty(storeID))
        {
          mapiRootFolder = outlookApp.Session.GetFolderFromID(entryID, storeID);
          items = !string.IsNullOrEmpty(filter) ? mapiRootFolder.Items.Restrict(filter)
            : mapiRootFolder.Items;

          Parallel.For(1, items.Count + 1, i =>
          {
            _MailItem mailItem = null;
            UserProperty userProperty = null;
            Outlook.Attachment attachment = null;

            try
            {

              mailItem = items.Item(i) as _MailItem;

              if (mailItem != null)
              {
                var mail = new Mail
                {
                  Body = mailItem.Body,
                  Subject = mailItem.Subject,
                  IsBodyHtml = GetMailBodyFormat(mailItem) == OutlookClient.OlBodyFormat.OlFormatHtml,
                  SentDateTime = mailItem.SentOn,
                  CreationDateTime = mailItem.CreationTime,
                  ID = mailItem.EntryID,
                  From = MailAddressMapper.Find(((dynamic)mailItem).SenderEmailAddress, outlookApp),
                };

                if (mail.IsBodyHtml)
                {
                  mail.Body = mailItem.HTMLBody;
                }

   
                for (int j = 1; j <= mailItem.Recipients.Count; j++)
                {
                  var recipient = mailItem.Recipients.Item(j);
                  switch (recipient.Type)
                  {
                    case 2:  // type cc = 2
                      mail.CC.Add(MailAddressMapper.Find(recipient.Address, outlookApp));
                      break;

                    case 3: // type bcc =3 
                      mail.BCC.Add(MailAddressMapper.Find(recipient.Address, outlookApp));
                      break;

                    default:  // type to = 1
                      mail.To.Add(MailAddressMapper.Find(recipient.Address, outlookApp));
                      break;
                  }
                }


                mails.Add(mail);
              }
            }
            catch (Exception) { }
            finally
            {
              SafeRelease(attachment);
              SafeRelease(mailItem);
              SafeRelease(userProperty);
            }
          });
        }

        EndSecurity(secMan, outlookApp);
      }
      catch (Exception) { }
      finally
      {
        SafeRelease(items);
        SafeRelease(mapiRootFolder);
        SafeRelease(outlookApp);
      }

      return mails;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="mailItem"></param>
    /// <returns></returns>
    private OlBodyFormat GetMailBodyFormat(_MailItem mailItem)
    {
      if (mailItem == null) return OlBodyFormat.OlFormatUnspecified;

      return (OlBodyFormat)mailItem.GetType().InvokeMember("BodyFormat",
        BindingFlags.GetProperty, null, mailItem, null);
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="obj"></param>
    private static void SafeRelease(object obj)
    {
      if (obj != null && Marshal.IsComObject(obj))
      {
        Marshal.ReleaseComObject(obj);
      }
      obj = null;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="outlookApp"></param>
    private SecurityManager BeginSecurity(Application outlookApp)
    {
      SecurityManager secMan = null;
      try
      {
        secMan = new SecurityManager();
        secMan.ConnectTo(outlookApp);
        secMan.DisableOOMWarnings = true;
        secMan.DisableSMAPIWarnings = true;
        secMan.DisableCDOWarnings = true;
      }
      catch (Exception ex)
      {
        throw ex;
      }
      return secMan;
    }
    /// <summary>
    /// 
    /// </summary>
    /// <param name="secMan"></param>
    /// <param name="outlookApp"></param>
    private void EndSecurity(SecurityManager secMan, Application outlookApp)
    {
      try
      {
        if (secMan != null)
        {
          secMan.DisableOOMWarnings = false;
          secMan.DisableSMAPIWarnings = false;
          secMan.DisableCDOWarnings = false;
          secMan.Disconnect(outlookApp);
        }
      }
      catch (Exception ex)
      {
        throw ex;
      }
    }

    public Folder CreateFolderByID(string entryID, string storeID, bool parent)
    {
      Folder folder = null;
      Application outlookApp = null;

      try
      {
        try
        {
          outlookApp = Marshal.GetActiveObject("Outlook.Application") as Application;
        }
        catch (Exception)
        {
          outlookApp = new Application();
        }

        var secMan = BeginSecurity(outlookApp);
        MAPIFolder mapiFolder = outlookApp.Session.GetFolderFromID(entryID, storeID);

        if (parent)
        {
          mapiFolder = mapiFolder.Parent as MAPIFolder;
        }

        if (mapiFolder != null)
        {
          folder = new Folder(this);
          folder.DisplayName = mapiFolder.Name;
          folder.Name = mapiFolder.Name;
          folder.ID = string.Format("{0}*{1}", mapiFolder.EntryID, mapiFolder.StoreID);
        }

        EndSecurity(secMan, outlookApp);
      }
      catch (Exception) { }
      finally
      {
        SafeRelease(outlookApp);
      }

      return folder;
    }


    #endregion
  }
}
